----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    18:07:09 11/11/2015 
-- Design Name: 
-- Module Name:    Datapath - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Datapath is
  PORT(
    clk_i : in std_logic;
    --Input
    data_i    : in std_logic_vector(7 downto 0);
    key_i     : in std_logic_vector(7 downto 0);
    --Control
    mode_i    : in std_logic;
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(2 downto 0);
    tmp_sel   : in std_logic;
    count_r   : in std_logic_vector(4 downto 0);
    --Output
    data_o    : out std_logic_vector(7 downto 0)
  );
end Datapath;

architecture Behavioral of Datapath is

  COMPONENT klein_keyschedule
	PORT(
    mode_i  : IN STD_LOGIC;
		key_in : IN std_logic_vector(0 to 79);
		i : IN std_logic_vector(4 downto 0);          
		key_out : OUT std_logic_vector(0 to 79)
		);
	END COMPONENT;
  
  --cipher
  COMPONENT klein_subnibbles
	PORT(
		byte_i : IN std_logic_vector(0 to 7);          
		byte_o : OUT std_logic_vector(0 to 7)
		);
	END COMPONENT;
  
  COMPONENT klein_mixnibbles
	PORT(
		h_state_in : IN std_logic_vector(0 to 31);          
		byte_out : OUT std_logic_vector(0 to 7)
		);
	END COMPONENT;
  
  COMPONENT klein_mixnibbles_d
	PORT(
		h_state_in : IN std_logic_vector(0 to 31);          
		byte_o : OUT std_logic_vector(0 to 7)
		);
	END COMPONENT;
  
  --signals
  --regsiters
  signal key_reg      : std_logic_vector(0 to key_size-1);
  signal state_reg    : std_logic_vector(0 to state_size-1);
  --process signals
  signal key_o        : std_logic_vector(0 to key_size-1);
  signal keyd_o       : std_logic_vector(0 to key_size-1);
  signal add_rk_i     : std_logic_vector(0 to 7);
  signal rk_i         : std_logic_vector(0 to 7);
  signal add_rk       : std_logic_vector(0 to 7);
  signal sub_i        : std_logic_vector(0 to 7);
  signal sub_o        : std_logic_vector(0 to 7);
  signal mixd_o_s     : std_logic_vector(0 to 7);
  signal tmp_reg      : std_logic_vector(0 to 31);
  signal round_o      : std_logic_vector(0 to 7);
  
begin
  --key process
  K_proc : process(clk_i, key_sel)
  begin
    if rising_edge(clk_i) then
      if key_sel = "01" then
        key_reg <= key_i & key_reg(0 to key_size-9);
      elsif key_sel = "10" then
        key_reg(0 to 63)  <= key_reg(56 to 63) & key_reg(0 to 55);
        key_reg(64 to 79) <= key_reg(64 to 79);
      elsif key_sel = "11" then
        key_reg <= key_o;
      else
        key_reg <= key_reg;
      end if;
    end if;
  end process;
  
  --data process
  D_proc : process(clk_i, data_sel)
  begin
    if rising_edge(clk_i) then
      if data_sel = "001" then
        state_reg <= data_i & state_reg(0 to state_size-9);
      elsif data_sel = "010" then
        state_reg <=  state_reg(56 to 63) & state_reg(0 to state_size-9);
      elsif data_sel = "011" then
        state_reg(0 to 15)  <=  state_reg(56 to 63) & state_reg(0 to 7);
        state_reg(16 to 47) <=  state_reg(16 to 47);
        state_reg(48 to 63) <=  round_o & state_reg(48 to 55);
      elsif data_sel = "100" then
        state_reg <= state_reg(16 to 47) & round_o & state_reg(48 to 63) & state_reg(0 to 7);
      elsif data_sel = "111" then
        state_reg(0 to 31)  <= sub_o & state_reg(0 to 23);
        state_reg(32 to 63) <= state_reg(32 to 63);
      elsif data_sel = "101" then
        state_reg <= state_reg(48 to 63) & sub_o & state_reg(0 to 23) & state_reg(32 to 47);
      else
        state_reg <= state_reg;
      end if;
    end if;
  end process;
  
  data_o <= state_reg(56 to 63) xor key_reg(56 to 63);
  
  --cipher component
  -- key
  Inst_klein_keyschedule: klein_keyschedule PORT MAP(
    mode_i => mode_i,
		key_in => key_reg,
		i => count_r,
		key_out => key_o
	);
  --data
  with mode_i select
    add_rk_i <= 
      state_reg(40 to 47) when '0',
      state_reg(56 to 63) when others;
  
  with mode_i select
    rk_i <= 
      key_reg(40 to 47) when '0',
      key_reg(56 to 63) when others;
    
  add_rk    <= add_rk_i xor rk_i;
  
  Inst_klein_subnibbles: klein_subnibbles PORT MAP(
		byte_i => sub_i,
		byte_o => sub_o
	);
  
  tmp_proc : process(clk_i, tmp_sel, mode_i)
  begin
    if rising_edge(clk_i) then
      if tmp_sel = '0' then
        if mode_i = '0' then
          tmp_reg(0 to 7) <= sub_o;
        else
          tmp_reg(0 to 7) <= add_rk;
        end if;
      else
        tmp_reg(0 to 7) <= tmp_reg(24 to 31);
      end if;
      tmp_reg(8 to 31)  <= tmp_reg(0 to 23);
    end if;
  end process;
  
  Inst_klein_mixnibbles: klein_mixnibbles PORT MAP(
		h_state_in => tmp_reg,
		byte_out   => round_o
	);
  
  Inst_klein_mixnibbles_d: klein_mixnibbles_d PORT MAP(
		h_state_in => tmp_reg,
		byte_o     => mixd_o_s
	);
  
  with mode_i select
    sub_i <=  
      add_rk   when '0',
      mixd_o_s when others;
  
end Behavioral;

