----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    18:33:09 10/28/2015 
-- Design Name: 
-- Module Name:    control_unit - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity control_unit is
  generic (
    state_size  : in integer := 64;
    key_size    : in integer := 80
  );
  port(
    clk_i     : in std_logic;
    rstb_i    : in std_logic;
    --instruction
    control_i : in std_logic_vector(1 downto 0);
    --output flag
    key_ready : out std_logic;
    busy_o    : out std_logic;
    done_o    : out std_logic;
    --control signal
    mode_o    : out std_logic;
    key_sel   : out std_logic_vector(1 downto 0);
    data_sel  : out std_logic_vector(2 downto 0);
    count_r   : out std_logic_vector(4 downto 0);
    tmp_sel   : out std_logic
  );
end control_unit;

architecture Behavioral of control_unit is
--subtype SM_type is std_logic_vector(3 downto 0);
--constant idle           : SM_type:="0000";
--constant load_key       : SM_type:="0001";
--constant ready          : SM_type:="0010";
--constant load_data_enc  : SM_type:="0011";
--constant load_data_dec  : SM_type:="0110";
--constant init           : SM_type:="0111";
--constant encryption     : SM_type:="1000";
--constant decryption     : SM_type:="1001";
--constant release        : SM_type:="1011";
-- state nachine states
type SM_type is (
  idle,
  ready,
  load_key,
  load_data_enc,
  encryption,
  load_data_dec,
  init,
  decryption,
  release
);

signal c_state : SM_type;
signal n_state : SM_type;

--internal signals
signal key_ready_s  : std_logic;
signal ena_c        : std_logic_vector(1 downto 0);
signal ena_r        : std_logic_vector(1 downto 0);
signal ena_ir        : std_logic;
--counters
signal load_count   : std_logic_vector(3 downto 0);
signal round_count  : std_logic_vector(4 downto 0);
signal intern_count  : std_logic_vector(3 downto 0);

begin

-- current state
cs_proc : process(rstb_i, clk_i)
begin
  if rstb_i = '0' then
    c_state <= idle;
  elsif rising_edge(clk_i) then
    c_state <= n_state;
  end if;
end process;

--next state
ns_proc : process(c_state, control_i, load_count, round_count)
begin
  n_state <= idle;
  case c_state is
  when idle =>
    if control_i = "01" then
      n_state <= load_key;
    end if;
  when load_key =>
    n_state <= load_key;
    if unsigned(load_count)=(key_size/8)-1 then
      n_state <= ready;
    end if;
  when ready =>
    n_state <= ready;
    if control_i = "01" then
      n_state <= load_key;
    elsif control_i = "10" then
      n_state <= load_data_enc;
    elsif control_i = "11" then
      n_state <= load_data_dec;
    end if;
  when load_data_enc =>
    n_state <= load_data_enc;
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= encryption;
    end if;
  when load_data_dec =>
    n_state <= load_data_dec;
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= init;
    end if;
  when init =>
    n_state <= init;
    if unsigned(round_count)=15 then
      n_state <= decryption;
    end if;
  when encryption =>
    n_state <= encryption;
    if unsigned(round_count)=15 and unsigned(load_count)=15 then
      n_state <= release;
    end if;
  when decryption =>
    n_state <= decryption;
    if unsigned(round_count)=1 and unsigned(load_count)=15 then
      n_state <= release;
    end if;
  when release =>
    n_state <= release;
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= idle;
    end if;
  when others =>
  end case;
end process;

--output
out_proc : process(c_state, load_count)
begin
  key_ready_s <= '0';
  busy_o      <= '0';
  mode_o      <= '0';
  done_o      <= '0';
  ena_c       <= "00";
  data_sel    <= "000";
  key_sel     <= "00";
  ena_r       <= "00";
  case c_state is
  when load_key =>
    busy_o  <= '1';
    ena_c       <= "01";
    key_sel <= "01";
  when ready =>
    key_ready_s <= '1';
  when load_data_enc =>
    busy_o      <= '1';
    ena_c       <= "10";
    data_sel    <= "001";
  when load_data_dec =>
    busy_o      <= '1';
    ena_c       <= "10";
    data_sel    <= "001";
  when init =>
    busy_o      <= '1';
    key_sel     <= "11";
    ena_r       <= "01";
  when encryption =>
    busy_o      <= '1';
    ena_c       <= "11";
    data_sel <= "010";
    ena_r    <= "11";
    if unsigned(load_count)<4 then 
      data_sel <= "010";
      key_sel  <= "10";
    elsif unsigned(load_count)<8 then 
      data_sel <= "011";
    elsif unsigned(load_count)<12 then 
      data_sel <= "010";
      key_sel     <= "10";
    elsif unsigned(load_count)<15 then 
      data_sel <= "011";
    else
      data_sel    <= "100";
      key_sel     <= "11";
      ena_r       <= "01";
    end if;
  when decryption =>
    busy_o      <= '1';
    mode_o      <= '1';
    ena_c       <= "11";
    data_sel    <= "010";
    ena_r       <= "11";
    if unsigned(load_count)<4 then 
      data_sel    <= "010";
      key_sel     <= "10";
    elsif unsigned(load_count)<8 then 
      data_sel    <= "111";
    elsif unsigned(load_count)<12 then 
      data_sel    <= "010";
      key_sel     <= "10";
    elsif unsigned(load_count)<15 then 
      data_sel <= "111";
    else
      data_sel    <= "101";
      key_sel     <= "11";
      ena_r       <= "10";
    end if;
  when release =>
    busy_o      <= '1';
    done_o      <= '1';
    ena_c       <= "10";
    data_sel    <= "001";
    key_sel <= "01";
  when others =>
  end case;
end process;

key_ready <= key_ready_s;

--load/release counter
L_count : process(rstb_i, clk_i, ena_c, load_count)
begin
  if rstb_i = '0' then
    load_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_c = "01" then
      if unsigned(load_count) < (key_size/8)-1 then
        load_count <= std_logic_vector(unsigned(load_count)+1);
      else
        load_count <= (others => '0');
      end if;
    elsif ena_c = "10" then
      if unsigned(load_count) < (State_size/8)-1 then
        load_count <= std_logic_vector(unsigned(load_count)+1);
      else
        load_count <= (others => '0');
      end if;
    elsif ena_c = "11" then
      load_count <= std_logic_vector(unsigned(load_count)+1);
    else
      load_count <= (others => '0');
    end if;
  end if;
end process;

R_count : process(rstb_i, clk_i, ena_r)
begin
  if rstb_i = '0' then
    round_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_r = "01" then
      round_count <= std_logic_vector(unsigned(round_count)+1);
    elsif ena_r = "10" then
      round_count <= std_logic_vector(unsigned(round_count)-1);
    elsif ena_r = "11" then
      round_count <= round_count;
    else
      round_count <= (others => '0');
    end if;
  end if;
end process;

tmp_sel <= load_count(2);
count_r <= round_count;

end Behavioral;

