--Klein, keyschedule

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY klein_keyschedule IS
  PORT (
    mode_i  : IN STD_LOGIC;
    key_in  : IN STD_LOGIC_VECTOR(0 to 79);
    i       : IN STD_LOGIC_VECTOR(4 DOWNTO 0);
    key_out : OUT STD_LOGIC_VECTOR(0 to 79)
    );
END klein_keyschedule;

ARCHITECTURE rtl OF klein_keyschedule IS

signal tmp : STD_LOGIC_VECTOR(0 to 79);

COMPONENT klein_sbox IS
  PORT (
    a_in  : IN STD_LOGIC_VECTOR(0 to 3);
    a_out : OUT STD_LOGIC_VECTOR(0 to 3)
    );
END COMPONENT;

signal sub_i1 : std_logic_vector(0 to 3);
signal sub_i2 : std_logic_vector(0 to 3);
signal sub_i3 : std_logic_vector(0 to 3);
signal sub_i4 : std_logic_vector(0 to 3);
signal sub_o1 : std_logic_vector(0 to 3);
signal sub_o2 : std_logic_vector(0 to 3);
signal sub_o3 : std_logic_vector(0 to 3);
signal sub_o4 : std_logic_vector(0 to 3);

signal tmp_add : std_logic_vector(0 to 4);



BEGIN
  
  with mode_i select tmp <= 
    (key_in(48 to 79) & key_in(40 to 47)) & 
    ((key_in(8 to 39) & key_in(0 to 7)) xor (key_in(48 to 79) & key_in(40 to 47))) when '0',
    key_in(40 to 47) & sub_o1 & sub_o2 & sub_o3 & sub_o4 & key_in(64 to 79) &
    key_in(0 to 18) & tmp_add & key_in(24 to 39) when others;

  sub_i1 <= tmp(48 to 51) when mode_i='0' else key_in(48 to 51);
  sub_i2 <= tmp(52 to 55) when mode_i='0' else key_in(52 to 55);
  sub_i3 <= tmp(56 to 59) when mode_i='0' else key_in(56 to 59);
  sub_i4 <= tmp(60 to 63) when mode_i='0' else key_in(60 to 63);

  i_klein_sbox1 : klein_sbox
    PORT MAP (
      a_in  => sub_i1,
      a_out => sub_o1 --key_out(48 to 51)
      );

  i_klein_sbox2 : klein_sbox
    PORT MAP (
      a_in  => sub_i2,
      a_out => sub_o2 --key_out(52 to 55)
      );
  
  i_klein_sbox3 : klein_sbox
    PORT MAP (
      a_in  => sub_i3,
      a_out => sub_o3 --key_out(56 to 59)
      );

   i_klein_sbox4 : klein_sbox
    PORT MAP (
      a_in  => sub_i4,
      a_out => sub_o4 --key_out(60 to 63)
      );
      
   with mode_i select
    tmp_add <= 
      tmp(19 to 23) xor std_logic_vector(unsigned(i)+1) when '0',
      key_in(19 to 23) xor i when others;
   
   with mode_i select key_out <=
    tmp(0 to 18) & tmp_add & tmp(24 to 39) & 
    tmp(40 to 47) & sub_o1 & sub_o2 & sub_o3 & sub_o4 & tmp(64 to 79) when '0',
    ((tmp(32 to 39) & tmp(0 to 31)) xor (tmp(72 to 79) & tmp(40 to 71))) &
    (tmp(72 to 79) & tmp(40 to 71)) when others;
  
END rtl;

