--Klein, mixnibbles step (decryption)

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY klein_mixnibbles_d IS
  PORT (
    h_state_in  : IN STD_LOGIC_VECTOR(0 to 31);
    byte_o : OUT STD_LOGIC_VECTOR(0 to 7)
    );
END klein_mixnibbles_d;

ARCHITECTURE rtl OF klein_mixnibbles_d IS

  COMPONENT klein_xtimes IS
    PORT (
      x_in  : IN STD_LOGIC_VECTOR(0 to 7);
      x_out : OUT STD_LOGIC_VECTOR(0 to 7)
      );
  END COMPONENT;
  
  signal xt1_in  : STD_LOGIC_VECTOR(0 to 7);
  signal xt2_in  : STD_LOGIC_VECTOR(0 to 7);
  signal xt3_in  : STD_LOGIC_VECTOR(0 to 7);
  signal xt1_out : STD_LOGIC_VECTOR(0 to 7);
  signal xt2_out : STD_LOGIC_VECTOR(0 to 7);
  signal xt3_out : STD_LOGIC_VECTOR(0 to 7);

BEGIN
  
  xt1_in <= h_state_in(0 to 7)    xor 
            h_state_in(8 to 15)   xor 
            h_state_in(16 to 23)  xor
            h_state_in(24 to 31);
  
  Inst_xtime1 : klein_xtimes port map(
    x_in  => xt1_in,
    x_out => xt1_out
  );
  
  xt2_in <= xt1_out             xor 
            h_state_in(24 to 31)  xor
            h_state_in(8 to 15);
            
  Inst_xtime2 : klein_xtimes port map(
    x_in  => xt2_in,
    x_out => xt2_out
  );
  
  xt3_in <= xt2_out             xor 
            h_state_in(24 to 31)  xor
            h_state_in(0 to 7);
            
  Inst_xtime3 : klein_xtimes port map(
    x_in  => xt3_in,
    x_out => xt3_out
  );
  
  byte_o <= xt3_out               xor 
            h_state_in(8 to 15)   xor 
            h_state_in(16 to 23)  xor
            h_state_in(0 to 7);

END rtl;

