----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    15:19:04 11/04/2015 
-- Design Name: 
-- Module Name:    round - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity round is
  port (
    clk_i   : in std_logic;
    rstb_i  : in std_logic;
    ena_ir  : in std_logic;
    byte_i  : in std_logic_vector(0 to 7);
    key_i   : in std_logic_vector(0 to 7);
    byte_o  : out std_logic_vector(0 to 7)
  );
end round;

architecture Behavioral of round is

  COMPONENT klein_subnibbles
	PORT(
		byte_i : IN std_logic_vector(0 to 7);          
		byte_o : OUT std_logic_vector(0 to 7)
		);
	END COMPONENT;
  
  COMPONENT klein_mixnibbles
	PORT(
		h_state_in : IN std_logic_vector(0 to 31);          
		byte_out : OUT std_logic_vector(0 to 7)
		);
	END COMPONENT;

  signal intern_count : std_logic_vector(0 to 2);
  
  signal add_rk       : std_logic_vector(0 to 7);
  signal sub_o        : std_logic_vector(0 to 7);
  signal tmp_reg      : std_logic_vector(0 to 31);

begin

  add_rk <= byte_i xor key_i;
  
  Inst_klein_subnibbles: klein_subnibbles PORT MAP(
		byte_i => add_rk,
		byte_o => sub_o
	);
  
  tmp_proc : process(clk_i, rstb_i, intern_count)
  begin
    if rstb_i = '0' then
      tmp_reg <= (others => '0');
    elsif rising_edge(clk_i) then
      if ena_ir = '0' then
        tmp_reg <= sub_o & tmp_reg(0 to 23);
      else
        tmp_reg <= tmp_reg(24 to 31) & tmp_reg(0 to 23);
      end if;
    end if;
  end process;
  
  Inst_klein_mixnibbles: klein_mixnibbles PORT MAP(
		h_state_in => tmp_reg,
		byte_out => byte_o
	);

end Behavioral;

